"""apply branch label

Revision ID: 90594e768102
Revises: 713e40249322
Create Date: 2025-07-17 15:45:43.718067

"""
from typing import Sequence

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '90594e768102'
down_revision: str | None = '713e40249322'
branch_labels: str | Sequence[str] | None = ('sc-audit',)
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



"""
Fetch sink events from Mercury Retroshades.

Author: Alex Olieman <https://keybase.io/alioli>
"""

import datetime as dt
from decimal import ROUND_DOWN, Decimal

import httpx
from pydantic import BaseModel
from stellar_sdk.sep.toid import TOID

from sc_audit.config import settings
from sc_audit.constants import KG, UNIT_IN_STROOPS


class SinkEvent(BaseModel):
    transaction: str
    created_at: dt.datetime
    contract_id: str
    funder: str
    recipient: str
    amount: Decimal
    project_id: str
    memo_text: str | None
    ledger: int

    @classmethod
    def from_raw(cls, **data):
        data["amount"] = (data["amount"] / UNIT_IN_STROOPS).quantize(KG, rounding=ROUND_DOWN)
        data["created_at"] = dt.datetime.fromtimestamp(data["timestamp"], dt.UTC)
        return cls(**data)


def get_sink_events(cursor: int=settings.FIRST_SINK_CURSOR) -> list[SinkEvent]:
    if not settings.MERCURY_KEY:
        raise MercuryError(
            "Skip sink events: Mercury key is not set in the configuration.",
            mercury_key=settings.MERCURY_KEY,
            retroshades_md5=settings.RETROSHADES_MD5,
        )
    elif settings.OBSRVR_FLOW_DB_URI:
        raise MercuryError(
            "Only one of MERCURY_KEY and OBSRVR_FLOW_DB_URI may be provided.",
            mercury_key=settings.MERCURY_KEY,
            retroshades_md5=settings.RETROSHADES_MD5,
        )

    headers = {
        "Authorization": settings.MERCURY_KEY,
        "Content-Type": "application/json",
    }
    ledger_from_toid = TOID.from_int64(cursor).ledger_sequence
    with httpx.Client(headers=headers) as client:
        payload = {
            "query": f"""
                SELECT * FROM sink_event{settings.RETROSHADES_MD5}
                WHERE ledger > {ledger_from_toid}
            """
        }
        resp: httpx.Response = client.post(
            url=str(settings.RETROSHADES_URL),
            json=payload,
        )
        if resp.is_error:
            err_msg = f"Failed to fetch sink events <{resp.status_code}>: {resp.text}"
            if resp.text.endswith('does not exist")'):
                err_msg = f"Retroshades table sink_event{settings.RETROSHADES_MD5} does not exist."
            raise MercuryError(
                err_msg,
                mercury_key=settings.MERCURY_KEY,
                retroshades_md5=settings.RETROSHADES_MD5,
            )
        
        events = resp.json()

    return [SinkEvent.from_raw(**event) for event in events]


class MercuryError(Exception):
    def __init__(self, msg, mercury_key: str, retroshades_md5: str):
        super().__init__(msg, mercury_key, retroshades_md5)



if __name__ == "__main__":
    sink_events = get_sink_events(cursor=0)
    pass



"""
Load sink invocations from sorocarbon into the DB as SinkingTx records.

Author: Alex Olieman <https://keybase.io/alioli>
"""

from sc_audit.config import settings
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.loader.sink_events import try_project_id
from sc_audit.session_manager import Session
from sc_audit.sources.sink_invocations import InvocationSource


def load_sink_invocations(cursor: int=settings.FIRST_SINK_CURSOR) -> int:
    """
    Load (all) sink invocations from Obsrvr Flow into the DB.

    To catch up with Flow, specify the cursor parameter to be the paging token
    of the latest sink invocation record present in the SinkingTx table.
    """
    number_loaded = 0

    with Session.begin() as session:
        for sink_invoke in InvocationSource.get_sink_invocations(cursor):
            vcs_project_id = try_project_id(sink_invoke.project_id)
            memo = sink_invoke.memo_text or None
            session.add(
                SinkingTx(
                    hash=sink_invoke.tx_hash,
                    created_at=sink_invoke.timestamp,
                    contract_id=sink_invoke.contract_id,
                    funder=sink_invoke.funder,
                    recipient=sink_invoke.recipient,
                    carbon_amount=sink_invoke.ton_amount,
                    source_asset_code=settings.CARBON_ASSET.code,
                    source_asset_issuer=settings.CARBON_ISSUER_PUB,
                    source_asset_amount=sink_invoke.ton_amount,
                    dest_asset_code=settings.CARBON_ASSET.code,
                    dest_asset_issuer=settings.CARBON_ISSUER_PUB,
                    dest_asset_amount=sink_invoke.ton_amount,
                    vcs_project_id=vcs_project_id,
                    memo_type='text' if memo else 'none',
                    memo_value=memo,
                    paging_token=sink_invoke.toid,
                )
            )
        
        number_loaded = len(session.new)

    return number_loaded



"""
Declare credit retirement DB models.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from __future__ import annotations
import datetime as dt
import typing

from sqlalchemy import ForeignKey, Index, String, event
from sqlalchemy.orm import Mapped, MappedAsDataclass, mapped_column, relationship

from sc_audit.db_schema.base import ScBase, intpk
from sc_audit.db_schema.impact_project import VcsProject
from sc_audit.loader.utils import VcsSerialNumber

if typing.TYPE_CHECKING:
    from sc_audit.db_schema import RetirementFromBlock, SinkStatus


InstrumentType = typing.Literal['VCU']


class RetirementBase(MappedAsDataclass, kw_only=True):
    __tablename__ = "retirements"

    certificate_id: Mapped[intpk]
    vcu_amount: Mapped[int]
    serial_number: Mapped[str] = mapped_column(String(128))
    retirement_date: Mapped[dt.date]
    retirement_beneficiary: Mapped[str]
    retirement_details: Mapped[str]
    vcs_project_id: Mapped[int] = mapped_column(ForeignKey("vcs_projects.id"))
    issuance_date: Mapped[dt.date]
    instrument_type: Mapped[InstrumentType]
    vintage_start: Mapped[dt.date]
    vintage_end: Mapped[dt.date]
    total_vintage_quantity: Mapped[int]


class Retirement(RetirementBase, ScBase):
    __tablename__ = "retirements"

    vcs_project: Mapped[VcsProject] = relationship(init=False, repr=False)
    retired_from: Mapped[list[RetirementFromBlock]] = relationship(
        init=False, repr=False, 
        back_populates='retirement',
    )
    sink_statuses: Mapped[list[SinkStatus]] = relationship(
        init=False, repr=False, 
        back_populates='retirement',
    )

    @property
    def tx_hashes_from_details(self) -> list[str]:
        tag, *tx_hashes = self.retirement_details.split()
        assert tag.startswith("stellarcarbon")
        return tx_hashes
    
    def validate_serial_number(self):
        value = self.serial_number
        if not value:
            raise ValueError("Serial number cannot be empty")
        
        serial_number = VcsSerialNumber.from_str(value)
        vcu_amount = 1 + serial_number.block_end - serial_number.block_start
        if serial_number.project_id != self.vcs_project_id:
            raise ValueError(
                f"Serial number {value} does not match VCS project ID {self.vcs_project_id}"
            )
        elif vcu_amount != self.vcu_amount:
            raise ValueError(
                f"Serial number {value} implies a VCU amount of {vcu_amount},"
                f" but got vcu_amount={self.vcu_amount}"
            )
    
    def as_dict(self, related: bool = False):
        data = {
            col: getattr(self, col)
            for col in self.__table__.columns.keys()
        }
        if related:
            del data['vcs_project_id']
            data['vcs_project'] = self.vcs_project.as_dict()
            data['retired_from'] = [
                rfb.as_dict() for rfb in self.retired_from
                if rfb.retirement_id
            ]
            data['sink_statuses'] = [
                status.as_dict() for status in self.sink_statuses
                if status.certificate_id
            ]
        
        return data


idx_retirement_date = Index("idx_retirement_date", Retirement.retirement_date.desc())
idx_retirement_beneficiary = Index("idx_retirement_beneficiary", Retirement.retirement_beneficiary)

@event.listens_for(Retirement, "before_insert")
def validate_retirement(mapper, connection, target: Retirement):
    target.validate_serial_number()



"""
Declare impact project DB models.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from __future__ import annotations
import typing

from sqlalchemy import Unicode
from sqlalchemy.orm import Mapped, MappedAsDataclass, mapped_column, relationship

from sc_audit.db_schema.base import ScBase, intpk

if typing.TYPE_CHECKING:
    from sc_audit.db_schema import MintedBlock


VcsCategory = typing.Literal["Agriculture Forestry and Other Land Use"]
VcsProtocol = typing.Literal["VM0015"]


class VcsProjectBase(MappedAsDataclass):

    id: Mapped[intpk]
    name: Mapped[str]
    category: Mapped[VcsCategory]
    protocol: Mapped[VcsProtocol]
    additional_certifications: Mapped[str | None]
    region: Mapped[str] = mapped_column(Unicode(128))
    country: Mapped[str] = mapped_column(Unicode(128))


class VcsProject(VcsProjectBase, ScBase):
    __tablename__ = "vcs_projects"

    minted_blocks: Mapped[list[MintedBlock]] = relationship(
        init=False, repr=False, 
        back_populates="vcs_project", 
        order_by="asc(MintedBlock.block_start)"
    )

    def as_dict(self):
        return {
            col: getattr(self, col)
            for col in self.__table__.columns.keys()
        }


class UnknownVcsProject(Exception):
    def __init__(self, msg, vcs_id: int):
        super().__init__(msg, vcs_id)


def get_vcs_project(vcs_project_id: int) -> VcsProject | None:
    vcs_projects = {
        1360: VcsProject(
            id=1360, 
            name="Forest Management to reduce deforestation and degradation in Shipibo Conibo and Cacataibo Indigenous communities of Ucayali region", 
            category="Agriculture Forestry and Other Land Use", 
            protocol="VM0015", 
            additional_certifications="CCB-Gold", 
            region="Latin America", 
            country="Peru"
        ),
    }
    return vcs_projects.get(vcs_project_id)



"""
Load the relationships between retirements and minted blocks into the DB.

These relationships are implicit in the source data. Both minted blocks and retirements
are associated with serial numbers that specify a range of individual credits. This loader
module iterates over retirement records and finds the minted blocks that match their serial
numbers, storing the explicit relationships.

Author: Alex Olieman <https://keybase.io/alioli>
"""

from sqlalchemy import and_, func, or_, select

from sc_audit.db_schema.association import RetirementFromBlock
from sc_audit.db_schema.mint import MintedBlock
from sc_audit.db_schema.retirement import Retirement
from sc_audit.loader.utils import VcsSerialNumber
from sc_audit.session_manager import Session


def load_retirement_from_block() -> int:
    """
    Load the retirement—block associations to ensure that all retirements are explicitly covered
    by their originating blocks.
    """
    number_loaded = 0

    with Session.begin() as session:
        # select the retirements that are not fully related to minted blocks
        query = (
            select(Retirement)
            .outerjoin(Retirement.retired_from)
            .group_by(Retirement.certificate_id)
            .having(
                func.coalesce(func.sum(RetirementFromBlock.vcu_amount), 0) < Retirement.vcu_amount
            )
        )
        uncovered_retirements = session.scalars(query).unique().all()
        for retirement in uncovered_retirements:
            from_blocks = cover_retirement(session, retirement)
            session.add_all(from_blocks)
            number_loaded += len(from_blocks)

    return number_loaded

def cover_retirement(session, retirement: Retirement) -> list[RetirementFromBlock]:
    serial_number = VcsSerialNumber.from_str(retirement.serial_number)
    q_matching_blocks = (
        select(MintedBlock).where(
            or_(
                # MintedBlock.block_start <= serial_number.block_start <= MintedBlock.block_end
                and_(
                    MintedBlock.block_start <= serial_number.block_start,
                    serial_number.block_start <= MintedBlock.block_end
                ),
                # MintedBlock.block_start <= serial_number.block_end <= MintedBlock.block_end
                and_(
                    MintedBlock.block_start <= serial_number.block_end,
                    serial_number.block_end <= MintedBlock.block_end
                ),
                # serial_number.block_start <= MintedBlock.block_start <= serial_number.block_end
                # serial_number.block_start <= MintedBlock.block_end <= serial_number.block_end
                and_(
                    MintedBlock.block_start.between(serial_number.block_start, serial_number.block_end),
                    MintedBlock.block_end.between(serial_number.block_start, serial_number.block_end)
                )
            )
        )
    )
    matching_blocks = session.scalars(q_matching_blocks).all()
    retirement_from_blocks = []
    for block in matching_blocks:
        rfb = RetirementFromBlock(
            retirement_id=retirement.certificate_id,
            block_hash=block.serial_hash,
            vcu_amount=overlap_size(serial_number, block)
        )
        raise_if_overspent(block, rfb)
        retirement_from_blocks.append(rfb)

    raise_if_uncovered(retirement, retirement_from_blocks)
    return retirement_from_blocks

def overlap_size(serial_number: VcsSerialNumber, block: MintedBlock) -> int:
    return max(
        0, (
            1 + min(serial_number.block_end, block.block_end) 
            - max(serial_number.block_start, block.block_start)
        )
    )


class BlockOverspent(Exception):
    def __init__(self, msg, block_hash: str, certificate_id: int):
        super().__init__(msg, block_hash, certificate_id)

def raise_if_overspent(block: MintedBlock, retirement_from_block: RetirementFromBlock):
    if retirement_from_block.vcu_amount > block.credits_remaining:
        raise BlockOverspent(
            f"Retiring {retirement_from_block.retirement_id} from block {block.serial_hash}"
            " would cause it to be overspent.",
            block_hash=block.serial_hash,
            certificate_id=retirement_from_block.retirement_id
        )
    

class CoveringBlockMissing(Exception):
    def __init__(self, msg, retirement: Retirement, new_blocks: list[RetirementFromBlock]):
        super().__init__(msg, retirement, new_blocks)

def raise_if_uncovered(retirement: Retirement, new_blocks: list[RetirementFromBlock]):
    new_coverage: int = sum(rfb.vcu_amount for rfb in new_blocks)
    if retirement.vcu_amount > new_coverage:
        raise CoveringBlockMissing(
            f"Retirement {retirement.certificate_id} is not fully covered by known blocks.",
            retirement=retirement,
            new_blocks=new_blocks
        )



"""
Helper functions to load only new records from data sources.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt
from typing import Literal, Union, overload

from sqlalchemy import select
from stellar_sdk.sep.toid import TOID

from sc_audit.config import settings
from sc_audit.db_schema.distribution import DistributionTx
from sc_audit.db_schema.mint import MintedBlock
from sc_audit.db_schema.retirement import Retirement
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.session_manager import Session


CoreModelName = Literal['dist_tx', 'mint_tx', 'retirement', 'sink_tx', 'sink_call']
LatestAttr = Union[dt.date, int, None]


@overload
def get_latest_attr(*models: *tuple[CoreModelName]) -> LatestAttr: ...

@overload
def get_latest_attr(*models: CoreModelName) -> list[LatestAttr]: ...

def get_latest_attr(*models: CoreModelName) -> LatestAttr | list[LatestAttr]:
    latest_attrs: list[LatestAttr] = [None] * len(models)
    with Session.begin() as session:
        for i, model in enumerate(models):
            if model == 'retirement':
                latest_retirement_date = session.scalar(
                    select(Retirement.retirement_date).order_by(Retirement.retirement_date.desc())
                )
                latest_attrs[i] = latest_retirement_date
            elif model == 'sink_tx':
                latest_sink_tx_cursor = session.scalar(
                    select(SinkingTx.paging_token)
                    .where(SinkingTx.contract_id.is_(None))
                    .order_by(SinkingTx.paging_token.desc())
                )
                latest_attrs[i] = increment_paging_token(latest_sink_tx_cursor) or settings.FIRST_SINK_CURSOR
            elif model == 'sink_call':
                latest_sink_call_cursor = session.scalar(
                    select(SinkingTx.paging_token)
                    .where(SinkingTx.contract_id.is_not(None))
                    .order_by(SinkingTx.paging_token.desc())
                )
                latest_attrs[i] = latest_sink_call_cursor or settings.FIRST_SINK_CURSOR
            elif model == 'mint_tx':
                latest_mint_tx_cursor = session.scalar(
                    select(MintedBlock.paging_token).order_by(MintedBlock.paging_token.desc())
                )
                latest_attrs[i] = increment_paging_token(latest_mint_tx_cursor) or settings.FIRST_MINT_CURSOR
            elif model == 'dist_tx':
                latest_dist_tx_cursor = session.scalar(
                    select(DistributionTx.paging_token).order_by(DistributionTx.paging_token.desc())
                )
                latest_attrs[i] = increment_paging_token(latest_dist_tx_cursor) or settings.FIRST_DIST_CURSOR

    if len(latest_attrs) == 1:
        return latest_attrs[0]
    else:
        return latest_attrs


def increment_paging_token(cursor: int | None) -> int | None:
    if cursor:
        toid = TOID.from_int64(cursor)
        toid.increment_operation_order()
        cursor = toid.to_int64()

    return cursor



"""
Load all database records in the order:
    1. Impact Projects
    2. Distribution Outflows
    3. Sinking Transactions
    4. Minted Blocks
    5. Retirements
    6. Retirement from Block
    7. Sink Statuses

Loading from scratch may fail if a Horizon instance with pruned history is selected.
You may select another instance with the SC_HORIZON_URL env variable.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt
import sys

from sc_audit.loader.distribution_outflows import load_distribution_txs
from sc_audit.loader.get_latest import get_latest_attr
from sc_audit.loader.impact_projects import load_impact_projects
from sc_audit.loader.minted_blocks import load_minted_blocks
from sc_audit.loader.retirement_from_block import load_retirement_from_block
from sc_audit.loader.retirements import load_retirements
from sc_audit.loader.sink_events import load_sink_events
from sc_audit.loader.sink_invocations import load_sink_invocations
from sc_audit.loader.sink_status import load_sink_statuses
from sc_audit.loader.sinking_txs import load_sinking_txs
from sc_audit.sources.sink_events import MercuryError
from sc_audit.sources.sink_invocations import ObsrvrError


def catch_up_from_sources():
    """
    Ensure that the latest available data is loaded from the original sources.
    Avoid loading duplicate records by filtering on cursor or date where possible.
    """
    retirement_date: dt.date
    sink_cursor: int
    contract_cursor: int
    mint_cursor: int
    dist_cursor: int
    retirement_date, sink_cursor, contract_cursor, mint_cursor, dist_cursor = get_latest_attr(
        'retirement', 
        'sink_tx', 
        'sink_call',
        'mint_tx',
        'dist_tx'
    ) # type: ignore[return-value]

    print("Started catch-up from data sources...")
    num_impact_projects = load_impact_projects()
    print(f"Loaded {num_impact_projects} impact projects")
    num_distribution_txs = load_distribution_txs(cursor=dist_cursor)
    print(f"Loaded {num_distribution_txs} distribution outflows")
    num_sinking_txs = load_sinking_txs(cursor=sink_cursor)
    print(f"Loaded {num_sinking_txs} sinking transactions")
    
    # attempt to load SinkContract txs from Obsrvr or Mercury
    try:
        num_sink_invocations = load_sink_invocations(cursor=contract_cursor)
        print(f"Loaded {num_sink_invocations} sink invocations")
    except ObsrvrError as exc:
        print(f"Couldn't load sink invocations from Obsrvr")
        print(repr(exc), file=sys.stderr)
        try:
            num_sink_events = load_sink_events(cursor=contract_cursor)
            print(f"Loaded {num_sink_events} sink events")
        except MercuryError as exc:
            print(f"Couldn't load sink events from Mercury")
            print(repr(exc), file=sys.stderr)
        
    num_minting_txs = load_minted_blocks(cursor=mint_cursor)
    print(f"Loaded {num_minting_txs} minted blocks")
    num_retirements = load_retirements(from_date=retirement_date)
    print(f"Loaded {num_retirements} retirements")
    num_retirement_from_block = load_retirement_from_block()
    print(f"Loaded {num_retirement_from_block} retirement-from-block associations")
    num_sink_statuses = load_sink_statuses()
    print(f"Loaded {num_sink_statuses} sink status associations")


if __name__ == "__main__":
    catch_up_from_sources()



from pathlib import Path

from alembic.config import Config
from alembic import command


package_root = Path(__file__).parent.parent
alembic_config = Config(package_root / "alembic.ini")


def current(verbose=False):
    command.current(alembic_config, verbose=verbose)


def upgrade(revision="head"):
    command.upgrade(alembic_config, revision)


def downgrade(revision):
    command.downgrade(alembic_config, revision)



"""
Load sinking transactions into the DB.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from decimal import Decimal
from typing import Any, Literal

from sc_audit.config import settings
from sc_audit.db_schema.impact_project import UnknownVcsProject, get_vcs_project
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.loader.utils import decode_hash_memo, parse_iso_datetime
from sc_audit.session_manager import Session
from sc_audit.sources.sinking_txs import get_sinking_transactions, get_tx_operations


def load_sinking_txs(cursor: int=settings.FIRST_SINK_CURSOR) -> int:
    """
    Load (all) sinking transactions from Horizon into the DB.

    To catch up with Horizon, specify the cursor parameter to be the incremented paging
    token of the latest record present in the DB. Always get the latest cursor once, and
    then load both sinking transactions and sink events using that cursor.
    """
    number_loaded = 0

    with Session.begin() as session:
        for sink_tx in get_sinking_transactions(cursor):
            # ensure that the related VCS Project is available
            vcs_project_id = get_vcs_project_id(sink_tx)
            vcs_project = get_vcs_project(vcs_project_id)
            if not vcs_project:
                raise UnknownVcsProject(
                    f"VCS project {vcs_project_id} needs to be loaded before related transactions"
                    " can be stored.",
                    vcs_id=vcs_project_id
                )
            
            tx_operations = get_tx_operations(sink_tx['transaction_hash'])
            payment_data = get_payment_data(tx_operations)
            memo_type = sink_tx['transaction']['memo_type']
            memo_value = sink_tx['transaction'].get('memo')
            if memo_type == 'hash':
                memo_value = decode_hash_memo(memo_value)

            session.add(
                SinkingTx(
                    hash=sink_tx['transaction_hash'],
                    created_at=parse_iso_datetime(sink_tx['created_at']),
                    contract_id=None,
                    funder=payment_data['funder'],
                    recipient=sink_tx['to'],
                    carbon_amount=Decimal(sink_tx['amount']),
                    source_asset_code=payment_data['source_asset_code'],
                    source_asset_issuer=payment_data['source_asset_issuer'],
                    source_asset_amount=Decimal(payment_data['source_asset_amount']),
                    dest_asset_code=payment_data['dest_asset_code'],
                    dest_asset_issuer=payment_data['dest_asset_issuer'],
                    dest_asset_amount=Decimal(payment_data['dest_asset_amount']),
                    vcs_project_id=vcs_project_id,
                    memo_type=memo_type,
                    memo_value=memo_value,
                    paging_token=sink_tx['paging_token'],
                )
            )
        
        number_loaded = len(session.new)

    return number_loaded


def get_vcs_project_id(sinking_tx) -> Literal[1360]:
    # TODO: support multiple VCS projects
    return 1360

def get_payment_data(transaction_operations: list[dict[str, Any]]) -> dict[str, Any]:
    for op in transaction_operations:
        # select the first payment operation
        if op['type_i'] in (1, 2, 13):
            source_asset_code = op.get('source_asset_code', op['asset_code'])
            source_asset_issuer = op.get('source_asset_issuer', op['asset_issuer'])
            if op.get('source_asset_type') == "native":
                source_asset_code = "XLM"
                source_asset_issuer = None

            return {
                'funder': op['from'],
                'source_asset_code': source_asset_code,
                'source_asset_issuer': source_asset_issuer,
                'source_asset_amount': op.get('source_amount', op['amount']),
                'dest_asset_code': op['asset_code'],
                'dest_asset_issuer': op['asset_issuer'],
                'dest_asset_amount': op['amount'],
            }
    else:
        raise ValueError("No payment operation found", transaction_operations)



"""
Load sink events from sorocarbon into the DB as SinkingTx records.

Author: Alex Olieman <https://keybase.io/alioli>
"""

from functools import lru_cache

from stellar_sdk.sep.toid import TOID

from sc_audit.config import settings
from sc_audit.db_schema.impact_project import get_vcs_project
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.session_manager import Session
from sc_audit.sources.sink_events import get_sink_events


def load_sink_events(cursor: int=settings.FIRST_SINK_CURSOR) -> int:
    """
    Load (all) sink events from Mercury Retroshades into the DB.

    To catch up with Retroshades, specify the cursor parameter to be the paging token
    of the latest sink event record present in the SinkingTx table.
    """
    number_loaded = 0

    with Session.begin() as session:
        last_ledger = 0
        # For each new ledger, reset tx_order to 2^16 to avoid collisions with SinkingTxs.
        # This is a pragmatic choice; ideally, the order should come from Retroshades.
        tx_order = first_tx_index = 2**16
        for sink_event in get_sink_events(cursor):
            vcs_project_id = try_project_id(sink_event.project_id)

            # FIXME: the TOID should come from Retroshades if possible
            if sink_event.ledger == last_ledger:
                tx_order += 1
            else:
                tx_order = first_tx_index

            event_toid = TOID(sink_event.ledger, tx_order, 0)
            last_ledger = sink_event.ledger
            # TODO: support multiple events per transaction
            # We can support >1 event per tx by abandoning tx hash as PK, replacing it
            # with the TOID (paging_token). Not done yet, because it requires a data migration.

            memo = sink_event.memo_text or None
            session.add(
                SinkingTx(
                    hash=sink_event.transaction,
                    created_at=sink_event.created_at,
                    contract_id=sink_event.contract_id,
                    funder=sink_event.funder,
                    recipient=sink_event.recipient,
                    carbon_amount=sink_event.amount,
                    source_asset_code=settings.CARBON_ASSET.code,
                    source_asset_issuer=settings.CARBON_ISSUER_PUB,
                    source_asset_amount=sink_event.amount,
                    dest_asset_code=settings.CARBON_ASSET.code,
                    dest_asset_issuer=settings.CARBON_ISSUER_PUB,
                    dest_asset_amount=sink_event.amount,
                    vcs_project_id=vcs_project_id,
                    memo_type='text' if memo else 'none',
                    memo_value=memo,
                    paging_token=event_toid.to_int64(),
                )
            )
        
        number_loaded = len(session.new)

    return number_loaded


@lru_cache(maxsize=32)
def try_project_id(project_id: str) -> int:
    """
    Try to parse the project_id into a VCS project ID.
    If it fails, return 1360 for the default VCS project.

    TODO: replace fallback with AnyProject so we can be more flexible with retirements.
    """
    if project_id.startswith("VCS"):
        try:
            vcs_id = int(project_id[3:])
            vcs_project = get_vcs_project(vcs_id)
            if vcs_project:
                return vcs_project.id
        except ValueError:
            pass

    return 1360  # Default VCS project ID



"""switch DistributionTx PK to paging_token

Revision ID: 52e5e68af772
Revises: 90594e768102
Create Date: 2025-11-10 19:49:34.286846

"""
from typing import Sequence

from alembic import context, op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '52e5e68af772'
down_revision: str | None = '90594e768102'
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def get_naming_convention():
    # get naming convention from the context
    migration_context = context.get_context()
    target_metadata = migration_context.opts.get("target_metadata")
    assert target_metadata is not None, "Target metadata must be set in the migration context"
    return target_metadata.naming_convention


def upgrade() -> None:
    naming_convention = get_naming_convention()

    with op.batch_alter_table('distribution_txs', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(None, type_='primary') # type: ignore
        batch_op.create_primary_key(batch_op.f('pk_distribution_txs'), ['paging_token'])
        batch_op.drop_index(batch_op.f('idx_dtx_toid'))
        batch_op.create_index('idx_dtx_hash', ['hash'], unique=False)

    with op.batch_alter_table('test_distribution_txs', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(None, type_='primary') # type: ignore
        batch_op.create_primary_key(batch_op.f('pk_test_distribution_txs'), ['paging_token'])


def downgrade() -> None:
    naming_convention = get_naming_convention()

    with op.batch_alter_table('distribution_txs', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(batch_op.f('pk_distribution_txs'), type_='primary')
        batch_op.create_primary_key(None, ['hash'])
        batch_op.drop_index('idx_dtx_hash')
        batch_op.create_index(batch_op.f('idx_dtx_toid'), ['paging_token'], unique=1)

    with op.batch_alter_table('test_distribution_txs', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint(batch_op.f('pk_test_distribution_txs'), type_='primary')
        batch_op.create_primary_key(None, ['hash'])



"""
SQLAlchemy Declarative Dataclass setup.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from dataclasses import fields
from decimal import Decimal
from typing import Annotated, Type

from sqlalchemy import Dialect, MetaData, String
import sqlalchemy.types as types
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import DeclarativeBase, MappedAsDataclass

from sc_audit.config import settings


class HexBinary(types.TypeDecorator):
    """
    Converts between fixed-length bytes and their hexadecimal string representations.
    """
    impl = types.LargeBinary
    cache_ok = True

    def process_bind_param(self, value: str, dialect: Dialect) -> bytes:
        return bytes.fromhex(value)
    
    def process_result_value(self, value: bytes, dialect: Dialect) -> str:
        if value is None:
            # this can happen during joins or eager loading
            return ""
        
        return value.hex()


intpk = Annotated[int, mapped_column(primary_key=True)]
txhash = Annotated[str, mapped_column(HexBinary(length=32))]
hashpk = Annotated[str, mapped_column(HexBinary(length=32), primary_key=True)]
strkey = Annotated[str, mapped_column(String(56))]
kgdecimal = Annotated[Decimal, mapped_column(types.DECIMAL(precision=21, scale=3))]
stroopdecimal = Annotated[Decimal, mapped_column(types.DECIMAL(precision=21, scale=7))]
bigint = Annotated[int, mapped_column(types.BigInteger())]
bigintpk = Annotated[int, mapped_column(types.BigInteger(), primary_key=True)]


class ScBase(MappedAsDataclass, DeclarativeBase):
    metadata = MetaData(naming_convention={
        "ix": "idx_%(column_0_name)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_`%(constraint_name)s`",
        "fk": "%(table_name)s_%(column_0_name)s_fkey",
        "pk": "pk_%(table_name)s"
    })

    def __init_subclass__(cls) -> None:
        prefix_tables_and_foreign_keys(cls)
        super().__init_subclass__()


def prefix_tables_and_foreign_keys(subclass: Type[DeclarativeBase]):
    if settings.TABLE_PREFIX:
        assert issubclass(subclass, MappedAsDataclass)
        # Check if the prefix has already been applied
        # unless the class is a test mapper class
        if (
            subclass.__name__.startswith('Test') 
            or not subclass.__tablename__.startswith(settings.TABLE_PREFIX)
        ):
            # Prefix the table name
            subclass.__tablename__ = f"{settings.TABLE_PREFIX}_{subclass.__tablename__}"

            for dfield in fields(subclass):
                field = getattr(subclass, dfield.name, None)
                if field and hasattr(field, 'foreign_keys'):
                    # Prefix all foreign key column references
                    for fk in field.foreign_keys:
                        if not fk._colspec.startswith(settings.TABLE_PREFIX):
                            fk._colspec = f"{settings.TABLE_PREFIX}_{fk._colspec}"
                        if fk.name and not fk.name.startswith(settings.TABLE_PREFIX):
                            fk.name = f"{settings.TABLE_PREFIX}_{fk.name}"


def create_test_mappers(base_class: Type[DeclarativeBase]):
    """
    Create testnet versions of all registered models.
    These mappers are not needed at runtime, but they must exist when generating migrations.
    """
    models = [
        mapper.class_
        for mapper in base_class.registry.mappers
    ]
    # TODO: generate indexes for test tables
    settings.TABLE_PREFIX = "test"
    for model in models:
        type(
            f'Test{model.__name__}', 
            model.__bases__, 
            {"__tablename__": model.__tablename__}
        )



"""add paging_token indexes

Revision ID: 713e40249322
Revises: 0c2f50c4d517
Create Date: 2025-06-26 14:13:08.992641

"""
from typing import Sequence

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '713e40249322'
down_revision: str | None = '0c2f50c4d517'
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    with op.batch_alter_table('distribution_txs', schema=None) as batch_op:
        batch_op.create_index('idx_dtx_toid', [sa.literal_column('paging_token DESC')], unique=True)  # type: ignore[arg-type]

    with op.batch_alter_table('minted_blocks', schema=None) as batch_op:
        batch_op.create_index('idx_block_toid', [sa.literal_column('paging_token DESC')], unique=True)  # type: ignore[arg-type]

    with op.batch_alter_table('sinking_txs', schema=None) as batch_op:
        batch_op.create_index('idx_stx_toid', [sa.literal_column('paging_token DESC')], unique=True)  # type: ignore[arg-type]


def downgrade() -> None:
    with op.batch_alter_table('sinking_txs', schema=None) as batch_op:
        batch_op.drop_index('idx_stx_toid')

    with op.batch_alter_table('minted_blocks', schema=None) as batch_op:
        batch_op.drop_index('idx_block_toid')

    with op.batch_alter_table('distribution_txs', schema=None) as batch_op:
        batch_op.drop_index('idx_dtx_toid')



"""
Fetch Stellarcarbon retirements from the Verra Registry.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt

import httpx
from parsel import Selector

from sc_audit.config import settings
from sc_audit.sources.common import verra_default_headers


def get_retirements_list(from_date: dt.date | None = None):
    with httpx.Client(headers=verra_default_headers) as client:
        query_filter = "startswith(retirementDetails,'stellarcarbon')"
        if from_date:
            iso_date = dt.datetime(from_date.year, from_date.month, from_date.day).isoformat()
            query_filter = f"({query_filter} and retireOrCancelDate ge {iso_date}.000Z)"

        headers = {
            "accept": "application/json",
            "referer": "https://registry.verra.org/app/search/VCS",
        }
        query_params = {
            "maxResults": 2000,
            "$filter": query_filter,
            "$count": True,
            "$orderby": "",
            "$skip": 0,
        }
        payload = {
            "program": "VCS",
            "assetStatus": "RETIRED",
            "issuanceTypeCodes": ["ISSUE"],
        }
        resp: httpx.Response = client.post(
            url=str(settings.VERRA_ASSET_SEARCH_URL),
            headers=headers,
            params=query_params,
            json=payload,
            timeout=settings.VERRA_ASSET_SEARCH_TIMEOUT
        )

    retirements_data = format_verra_retirements(resp.text)
    total_amount_retired = sum(
        rtm['vcu_amount'] 
        for rtm in retirements_data['retirements']
    )

    return {'total_amount_retired': total_amount_retired, **retirements_data}


def format_verra_retirements(retirements_payload: str):
    root = Selector(text=retirements_payload)
    total_count: int = int(root.jmespath('totalCount').get(default='0'))
    count_exceeded: bool = True if root.jmespath('countExceeded').get() == 'true' else False
    item_multiselect = """
        value[].{
            issuance_date: issuanceDate, 
            instrument_type: instrumentType,
            vintage_start: vintageStart,
            vintage_end: vintageEnd,
            total_vintage_quantity: totalVintageQuantity,
            vcs_id: resourceIdentifier,
            vcs_name: resourceName,
            vcs_category: protocolCategory,
            vcs_protocol: protocol,
            additional_certifications: additionalCertifications,
            region: region,
            country: country,
            vcu_amount: quantity,
            serial_numbers: serialNumbers,
            retirement_date: retireOrCancelDate,
            retirement_beneficiary: retirementBeneficiary,
            retirement_details: retirementDetails,
            certificate_id: holdingIdentifier
        }
    """
    items = root.jmespath(item_multiselect).getall()

    return {
        'total_count': total_count,
        'count_exceeded': count_exceeded,
        'retirements': items,
    }



"""
Fetch minting transactions from Stellar Horizon.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from stellar_sdk import Server
from stellar_sdk.server import PaymentsCallBuilder

from sc_audit.config import settings

server = Server(horizon_url=str(settings.HORIZON_URL))


def get_minting_transactions(cursor: int=settings.FIRST_MINT_CURSOR):
    query: PaymentsCallBuilder = (
        server.payments()
        .for_account(settings.CARBON_ISSUER_PUB)
        .cursor(cursor)
        .join("transactions")
        .order(desc=False)
        .limit(200)
    )
    resp = query.call()
    yield from filter_minting_txs(resp['_embedded']['records'])
    while next_records := query.next()['_embedded']['records']:
        yield from filter_minting_txs(next_records)

def filter_minting_txs(horizon_records):
    for payment in horizon_records:
        if (
            payment['from'] == settings.CARBON_ISSUER_PUB
            and payment['to'] == settings.CARBON_DISTRIB_PUB
            and payment['asset_code'] == settings.CARBON_ASSET.code
        ):
            yield payment


def get_carbon_outflows(cursor: int=settings.FIRST_DIST_CURSOR):
    query: PaymentsCallBuilder = (
        server.payments()
        .for_account(settings.CARBON_DISTRIB_PUB)
        .cursor(cursor)
        .order(desc=False)
        .limit(200)
    )
    resp = query.call()
    yield from filter_distribution_txs(resp['_embedded']['records'])
    while next_records := query.next()['_embedded']['records']:
        yield from filter_distribution_txs(next_records)

def filter_distribution_txs(horizon_records):
    for payment in horizon_records:
        if (
            payment["type_i"] == 1
            and payment["from"] == settings.CARBON_DISTRIB_PUB
            and payment["asset_code"] == settings.CARBON_ASSET.code
            and payment["asset_issuer"] == settings.CARBON_ASSET.issuer
        ):
            yield payment



"""
Download backups of DB tables from GitHub.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from base64 import b64decode
from importlib import metadata
from pathlib import Path
import tempfile

from github import Github
from github.GithubObject import NotSet, Opt


github = Github()
repo_name = "stellarcarbon/sc-data"


def download_compatible_dumps() -> Path:
    data_repo = github.get_repo(repo_name)
    package_version = metadata.version("sc_audit")
    matching_refs = data_repo.get_git_matching_refs("tags/sc-audit")
    for ref_obj in matching_refs:
        ref_version = ref_obj.ref.lstrip("refs/tags/sc-audit-v")
        if package_version.startswith(ref_version):
            # tagged data indicates latest compatible version
            return download_dumps(ref=ref_obj.ref)
        
    # it's safe to get the most recent backup
    return download_dumps()


def download_dumps(ref: Opt[str] = NotSet) -> Path:
    """
    Download DB table dump files from GitHub and return the tempdir where they are located.
    The caller of this function is responsible for deleting the tempdir after loading the files.
    """
    data_repo = github.get_repo(repo_name)
    # note: the `get_contents` endpoint will fail for files >100 MB
    dump_cfiles = data_repo.get_contents("sc-audit", ref=ref)
    assert isinstance(dump_cfiles, list)
    temp_dir = Path(tempfile.mkdtemp())
    for cfile in dump_cfiles:
        if cfile.encoding == "base64":
            # file size =< 1 MB, content is loaded
            content_bytes = cfile.decoded_content
        else:
            # file size =< 100 MB, need to fetch content
            git_blob = data_repo.get_git_blob(cfile.sha)
            content_bytes = b64decode(git_blob.content)

        write_path = temp_dir / cfile.name
        with write_path.open('wb') as dump_file:
            dump_file.write(content_bytes)
        
    return temp_dir



verra_default_headers = {
    "accept-encoding": "gzip, deflate, br",
    "accept-language": "en-US,en;q=0.5",
    "user-agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0",
}



"""
Declare CARBON issuance DB models.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from __future__ import annotations
from dataclasses import dataclass
import datetime as dt
import typing

from sqlalchemy import ForeignKey, Index, SQLColumnExpression, String, func, select
from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property
from sqlalchemy.orm import Mapped, MappedAsDataclass, mapped_column, relationship

from sc_audit.db_schema.base import HexBinary, ScBase, bigint, hashpk
from sc_audit.db_schema.impact_project import VcsProject

if typing.TYPE_CHECKING:
    from sc_audit.db_schema import RetirementFromBlock


VerraSubAccountName = typing.Literal[
    "CARBON Pool | stellarcarbon.io", 
    "CARBON Sink | stellarcarbon.io"
]

@dataclass
class VerraSubAccount:
    id: int
    name: VerraSubAccountName

verra_carbon_pool = VerraSubAccount(11273, "CARBON Pool | stellarcarbon.io")
verra_carbon_sink = VerraSubAccount(11274, "CARBON Sink | stellarcarbon.io")


class MintedBlockBase(MappedAsDataclass, kw_only=True):

    serial_hash: Mapped[hashpk]
    transaction_hash: Mapped[str] = mapped_column(HexBinary(length=32))
    created_at: Mapped[dt.datetime]
    vcs_project_id: Mapped[int] = mapped_column(ForeignKey('vcs_projects.id'))
    serial_number: Mapped[str] = mapped_column(String(128))
    block_start: Mapped[int]
    block_end: Mapped[int]
    sub_account_id: Mapped[int]
    sub_account_name: Mapped[VerraSubAccountName]
    vintage_start: Mapped[dt.date]
    vintage_end: Mapped[dt.date]
    paging_token: Mapped[bigint]


class MintedBlock(MintedBlockBase, ScBase):
    __tablename__ = "minted_blocks"

    vcs_project: Mapped[VcsProject] = relationship(init=False, repr=False, back_populates='minted_blocks')
    consumed_by: Mapped[list[RetirementFromBlock]] = relationship(
        init=False, repr=False, 
        back_populates='block',
        lazy='selectin',
        order_by="asc(RetirementFromBlock.retirement_id)"
    )

    @hybrid_property
    def size(self) -> int:
        return 1 + self.block_end - self.block_start
    
    @hybrid_property
    def credits_remaining(self) -> int:
        credits_consumed = sum(rfb.vcu_amount for rfb in self.consumed_by)
        return self.size - credits_consumed
    
    @credits_remaining.inplace.expression
    @classmethod
    def _credits_remaining_expression(cls) -> SQLColumnExpression[int]:
        from sc_audit.db_schema import RetirementFromBlock
        return (
            select(cls.size - func.coalesce(func.sum(RetirementFromBlock.vcu_amount), 0))
            .where(RetirementFromBlock.block_hash == cls.serial_hash)
            .label("credits_remaining")
        )
    
    @hybrid_method
    def credits_remaining_on_date(self, on_date: dt.date) -> int:
        credits_consumed = sum(
            rfb.vcu_amount
            for rfb in self.consumed_by
            if rfb.retirement.retirement_date <= on_date
        )
        return self.size - credits_consumed
    
    @credits_remaining_on_date.inplace.expression
    @classmethod
    def _credits_remaining_on_date_expression(cls, on_date: dt.date) -> SQLColumnExpression[int]:
        from sc_audit.db_schema import Retirement, RetirementFromBlock
        return (
            select(cls.size - func.coalesce(func.sum(RetirementFromBlock.vcu_amount), 0))
            .where(RetirementFromBlock.block_hash == cls.serial_hash)
            .where(RetirementFromBlock.retirement_id == Retirement.certificate_id)
            .where(Retirement.retirement_date <= on_date)
            .label("credits_remaining_on_date")
        )


idx_created_at = Index("idx_block_created_at", MintedBlock.created_at)
idx_toid = Index("idx_block_toid", MintedBlock.paging_token.desc(), unique=True)



"""remove retirement_beneficiary length constraint

Revision ID: 04d168a1598f
Revises: 4813b9c06c97
Create Date: 2024-08-12 12:32:56.638424

"""
from typing import Sequence

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '04d168a1598f'
down_revision: str | None = '4813b9c06c97'
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    with op.batch_alter_table('retirements', schema=None) as batch_op:
        batch_op.alter_column('retirement_beneficiary',
               existing_type=sa.VARCHAR(length=56),
               type_=sa.String(),
               existing_nullable=False)

    with op.batch_alter_table('test_retirements', schema=None) as batch_op:
        batch_op.alter_column('retirement_beneficiary',
               existing_type=sa.VARCHAR(length=56),
               type_=sa.String(),
               existing_nullable=False)


def downgrade() -> None:
    with op.batch_alter_table('test_retirements', schema=None) as batch_op:
        batch_op.alter_column('retirement_beneficiary',
               existing_type=sa.String(),
               type_=sa.VARCHAR(length=56),
               existing_nullable=False)

    with op.batch_alter_table('retirements', schema=None) as batch_op:
        batch_op.alter_column('retirement_beneficiary',
               existing_type=sa.String(),
               type_=sa.VARCHAR(length=56),
               existing_nullable=False)



"""
Load distribution outflows into the DB.

Author: Alex Olieman <https://keybase.io/alioli>
"""

from decimal import Decimal

from sc_audit.config import settings
from sc_audit.db_schema.distribution import DistributionTx
from sc_audit.loader.utils import parse_iso_datetime
from sc_audit.session_manager import Session
from sc_audit.sources.minting_txs import get_carbon_outflows


def load_distribution_txs(cursor: int=settings.FIRST_DIST_CURSOR) -> int:
    """
    Load (all) distribution outflows from Horizon into the DB.

    To catch up with Horizon, specify the cursor parameter to be the incremented paging
    token of the latest record present in the DB.
    """
    number_loaded = 0

    with Session.begin() as session:
        for payment in get_carbon_outflows(cursor):
            session.add(
                DistributionTx(
                    hash=payment['transaction_hash'],
                    created_at=parse_iso_datetime(payment['created_at']),
                    sender=payment['from'],
                    recipient=payment['to'],
                    carbon_amount=Decimal(payment['amount']),
                    paging_token=payment['paging_token'],
                )
            )
        
        number_loaded = len(session.new)

    return number_loaded



from .base import ScBase
from .association import RetirementFromBlock, SinkStatus
from .distribution import DistributionTx
from .impact_project import VcsProject
from .mint import MintedBlock
from .retirement import Retirement
from .sink import SinkingTx



"""
View Stellarcarbon's asset stats. Optionally, filter by recipient.

Author: Alex Olieman <https://keybase.io/alioli>
"""

from decimal import Decimal

from sqlalchemy import func, select, union_all

from sc_audit.db_schema.association import SinkStatus
from sc_audit.db_schema.mint import MintedBlock
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.session_manager import Session


def get_carbon_stats(recipient: str | None = None) -> dict[str, Decimal]:
    sel_sunk = select(func.sum(SinkingTx.carbon_amount))
    sel_retired = select(func.sum(SinkStatus.amount_filled))

    if recipient:
        sel_sunk = sel_sunk.where(SinkingTx.recipient == recipient)
        sel_retired = (
            sel_retired
            .join(SinkStatus.sinking_transaction)
            .where(SinkingTx.recipient == recipient)
        )
        q_stats = union_all(sel_sunk, sel_retired)
    else:
        sel_stored = select(func.sum(MintedBlock.size))
        q_stats = union_all(sel_sunk, sel_retired, sel_stored)

    with Session.begin() as session:
        db_res = (val or Decimal() for val in session.scalars(q_stats).all())

    stats = {
        "carbon_sunk": Decimal(),
        "carbon_retired": Decimal(),
        "carbon_pending": Decimal(),
    }
    if recipient:
        c_sunk, c_retired = db_res
    else:
        c_sunk, c_retired, c_stored = db_res
        stats["carbon_stored"] = c_stored

    stats["carbon_sunk"] = c_sunk
    stats["carbon_retired"] = c_retired
    stats["carbon_pending"] = c_sunk - c_retired

    return stats



"""
Load impact projects into the DB.

Author: Alex Olieman <https://keybase.io/alioli>
"""

from sqlalchemy import select
from sc_audit.db_schema.base import intpk
from sc_audit.db_schema.impact_project import VcsProject, get_vcs_project
from sc_audit.session_manager import Session


def load_impact_projects() -> int:
    """
    Load impact projects from our own data schema into the DB.
    """
    number_loaded = 0

    with Session.begin() as session:
        existing_vcs_projects: set[intpk] = set(session.scalars(select(VcsProject.id)).all())
        if 1360 not in existing_vcs_projects:
            vcs_1360 = get_vcs_project(1360)
            session.add(vcs_1360)
        
        number_loaded = len(session.new)

    return number_loaded



"""
View sinking transactions and their retirement status.

Transactions can be filtered by account, by date, and by their retirement status.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt
from typing import Any, Literal

import pandas as pd
from sqlalchemy import Select, exists, not_, select
from sqlalchemy.orm import contains_eager, selectinload

from sc_audit.db_schema.association import SinkStatus
from sc_audit.db_schema.sink import SinkingTx
from sc_audit.session_manager import Session


def view_sinking_txs(
        for_funder: str | None = None,
        for_recipient: str | None = None, 
        from_date: dt.date | None = None,
        before_date: dt.date | None = None,
        finalized: bool | None = None,
        contract_call: bool | None = None,
        cursor: int | None = None,
        limit: int | None = None,
        order: Literal['asc', 'desc'] = 'desc',
) -> pd.DataFrame:
    base_query = construct_stx_query(
        for_funder, for_recipient, from_date, before_date, finalized, contract_call
    ).subquery()
    subq_paging_token = base_query.c.paging_token
    # Determine order and cursor filter
    if order == 'asc':
        subq_order = subq_paging_token.asc()
        stx_order = SinkingTx.paging_token.asc()
        cursor_filter = subq_paging_token > (cursor or 0)
    elif order == 'desc':
        subq_order = subq_paging_token.desc()
        stx_order = SinkingTx.paging_token.desc()
        cursor_filter = subq_paging_token < (cursor or 0)

    with Session.begin() as session:
        # Step 1: Get unique paging_tokens for the page
        token_query = select(subq_paging_token).select_from(base_query)
        if cursor and cursor > 0:
            token_query = token_query.where(cursor_filter)
        
        token_query = token_query.order_by(subq_order).distinct().limit(limit)
        paging_tokens = session.scalars(token_query).all()

        if not paging_tokens:
            return pd.DataFrame()

        # Step 2: Fetch SinkingTxs for those paging_tokens, eager load statuses
        stx_query = (
            select(SinkingTx)
            .outerjoin(SinkStatus)
            .options(contains_eager(SinkingTx.statuses))
            .where(SinkingTx.paging_token.in_(paging_tokens))
            .order_by(stx_order)
        )
        stx_records = session.scalars(stx_query).unique().all()
        txdf = pd.DataFrame.from_records(stx.as_dict() for stx in stx_records)

    return txdf


def construct_stx_query(
        for_funder: str | None,
        for_recipient: str | None, 
        from_date: dt.date | None,
        before_date: dt.date | None,
        finalized: bool | None,
        contract_call: bool | None,
) -> Select[tuple[SinkingTx]]:
    # Only build the base query and filters
    q_txs = select(SinkingTx)
    if for_funder:
        q_txs = q_txs.where(SinkingTx.funder == for_funder)
    
    if for_recipient:
        q_txs = q_txs.where(SinkingTx.recipient == for_recipient)
    
    if from_date:
        from_dt = dt.datetime(from_date.year, from_date.month, from_date.day)
        q_txs = q_txs.where(SinkingTx.created_at >= from_dt)
    
    if before_date:
        before_dt = dt.datetime(before_date.year, before_date.month, before_date.day)
        q_txs = q_txs.where(SinkingTx.created_at < before_dt)
    
    if finalized is False:
        # filter on sink txs without any SinkStatus with finalized=True
        # this clause needs to be the perfect negation of `finalized is True`
        q_txs = q_txs.where(
            not_(exists().where(
                SinkStatus.sinking_tx_hash == SinkingTx.hash, SinkStatus.finalized == True
            ).correlate(SinkingTx))
        )
    elif finalized is True:
        # select sink txs that have at least one SinkStatus with finalized=True
        q_txs = q_txs.outerjoin(SinkStatus).where(SinkStatus.finalized == True)

    if contract_call is True:
        q_txs = q_txs.where(SinkingTx.contract_id.is_not(None))
    elif contract_call is False:
        q_txs = q_txs.where(SinkingTx.contract_id.is_(None))
    
    return q_txs


def get_sinking_tx(hash: str) -> dict[str, Any] | None:
    stx_query = (
        select(SinkingTx)
        .options(selectinload(SinkingTx.statuses))
        .where(SinkingTx.hash == hash)
    )
    with Session.begin() as session:
        stx = session.scalar(stx_query)
        stx_data = stx.as_dict() if stx else None

    return stx_data



"""
Fetch sink_carbon invocations from Obsrver Flow.

Author: Alex Olieman <https://keybase.io/alioli>
"""

import datetime as dt
from decimal import ROUND_DOWN, Decimal
from typing import Sequence

from sqlalchemy import URL, Engine, create_engine, make_url, select
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, Session

from sc_audit.config import settings
from sc_audit.constants import KG, UNIT_IN_STROOPS
from sc_audit.db_schema.base import bigint, intpk


class FlowBase(MappedAsDataclass, DeclarativeBase): ...


class SinkInvocation(FlowBase):
    __tablename__ = settings.OBSRVR_FLOW_TABLE

    id: Mapped[intpk]
    toid: Mapped[bigint]
    ledger: Mapped[int]
    timestamp: Mapped[dt.datetime]
    contract_id: Mapped[str]
    function_name: Mapped[str]
    invoking_account: Mapped[str]
    tx_hash: Mapped[str]

    funder: Mapped[str]
    recipient: Mapped[str]
    amount: Mapped[bigint]
    project_id: Mapped[str]
    memo_text: Mapped[str | None]

    processed_at: Mapped[dt.datetime]
    schema_name: Mapped[str]
    successful: Mapped[bool]
    created_at: Mapped[dt.datetime]

    @hybrid_property
    def ton_amount(self) -> Decimal:
        return (self.amount / UNIT_IN_STROOPS).quantize(KG, rounding=ROUND_DOWN)
    

class ObsrvrError(Exception):
    def __init__(self, msg, dsn_uri: str | None):
        super().__init__(msg, dsn_uri)
    

class InvocationSource:
    db_url: URL | None
    engine: Engine | None

    if not settings.OBSRVR_FLOW_DB_URI:
        db_url = None
        engine = None
    elif settings.MERCURY_KEY:
        raise ObsrvrError(
            "Only one of MERCURY_KEY and OBSRVR_FLOW_DB_URI may be provided.",
            dsn_uri=str(settings.OBSRVR_FLOW_DB_URI)
        )
    else:
        db_url = make_url(str(settings.OBSRVR_FLOW_DB_URI))
        engine = create_engine(db_url)

    @classmethod
    def get_sink_invocations(cls, cursor: int=settings.FIRST_SINK_CURSOR) -> Sequence[SinkInvocation]:
        if not cls.engine:
            raise ObsrvrError(
                "Skip sink invocations: Obsrvr DB URI is not set in the configuration.",
                dsn_uri=None
            )

        query = select(SinkInvocation).where(SinkInvocation.toid > cursor)
        
        with Session(cls.engine, expire_on_commit=False) as session:
            invokes = session.scalars(query).all()

        return invokes


if __name__ == "__main__":
    sink_events = InvocationSource.get_sink_invocations(cursor=0)
    pass




import pandas as pd


def format_df(df: pd.DataFrame, format: str) -> str:
    match format:
        case "csv":
            return df.to_csv()
        case "json":
            return df.to_json(orient='records', date_format='iso', date_unit='s', indent=2)
        case _:
            return df.to_string(show_dimensions=True)



"""
Declare carbon sinking DB models.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from __future__ import annotations
import datetime as dt
from  decimal import Decimal
import typing

from sqlalchemy import ForeignKey, Index, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, MappedAsDataclass, mapped_column, relationship

from sc_audit.db_schema.base import ScBase, bigint, hashpk, kgdecimal, strkey, stroopdecimal
from sc_audit.db_schema.impact_project import VcsProject

if typing.TYPE_CHECKING:
    from sc_audit.db_schema import SinkStatus


MemoType = typing.Literal['text', 'hash', 'none']


class SinkingTxBase(MappedAsDataclass, kw_only=True):

    hash: Mapped[hashpk]
    created_at: Mapped[dt.datetime]
    contract_id: Mapped[strkey | None]
    funder: Mapped[strkey]
    recipient: Mapped[strkey]
    carbon_amount: Mapped[kgdecimal]
    source_asset_code: Mapped[str] = mapped_column(String(12))
    source_asset_issuer: Mapped[strkey | None]
    source_asset_amount: Mapped[stroopdecimal]
    dest_asset_code: Mapped[str] = mapped_column(String(12))
    dest_asset_issuer: Mapped[strkey]
    dest_asset_amount: Mapped[stroopdecimal]
    vcs_project_id: Mapped[int] = mapped_column(ForeignKey("vcs_projects.id"))
    memo_type: Mapped[MemoType]
    memo_value: Mapped[str | None] = mapped_column(String(64))
    paging_token: Mapped[bigint]


class SinkingTx(SinkingTxBase, ScBase):
    __tablename__ = "sinking_txs"

    vcs_project: Mapped[VcsProject] = relationship(init=False, repr=False)
    statuses: Mapped[list[SinkStatus]] = relationship(
        init=False, repr=False, 
        back_populates='sinking_transaction',
        lazy='selectin',
        order_by="asc(SinkStatus.certificate_id)"
    )

    @hybrid_property
    def total_filled(self) -> Decimal:
        return sum(
            (ss.amount_filled for ss in self.statuses),
            start=Decimal()
        )
    
    def as_dict(self):
        data = {
            col: getattr(self, col)
            for col in self.__table__.columns.keys()
        }
        data['statuses'] = [
            status.as_dict() for status in self.statuses
            if status.certificate_id
        ]
        return data
    

idx_created_at = Index("idx_stx_created_at", SinkingTx.created_at.desc())
idx_toid = Index("idx_stx_toid", SinkingTx.paging_token.desc(), unique=True)
idx_funder = Index("idx_stx_funder", SinkingTx.funder)
idx_recipient = Index("idx_stx_recipient", SinkingTx.recipient)



"""add contract_id to sinking_txs

Revision ID: 77ea86674a18
Revises: 04d168a1598f
Create Date: 2025-06-10 14:32:14.335259

"""
from typing import Sequence

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '77ea86674a18'
down_revision: str | None = '04d168a1598f'
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    with op.batch_alter_table('sinking_txs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('contract_id', sa.String(length=56), nullable=True))

    with op.batch_alter_table('test_sinking_txs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('contract_id', sa.String(length=56), nullable=True))


def downgrade() -> None:
    with op.batch_alter_table('test_sinking_txs', schema=None) as batch_op:
        batch_op.drop_column('contract_id')

    with op.batch_alter_table('sinking_txs', schema=None) as batch_op:
        batch_op.drop_column('contract_id')
